/* realmem - Demonstrate techniques to allow extra memory access *
 *           in real-mode applications.  -James K. Lawless       */

#include <dos.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>

#ifdef __BORLANDC__
    #undef  FP_SEG
    #define FP_SEG(fp) (*((unsigned __far *)&(fp)+1))
#endif

typedef unsigned short WORD;

/* register work areas for int86() calls    */
union REGS regs;
struct SREGS sregs;

/* main test routines                       */
void VideoTest(void);
void UMBTest(void);
void HMATest(void);

/* prototypes for Video RAM routines        */
void VideoRAM(void);
void VMsg(WORD,char *);
void VideoNorm(void);

/* XMS routines                             */
WORD InitXMS(void);
WORD XMSAllocHMA(void);
WORD XMSEnableA20(void);
WORD XMSDisableA20(void);

/* pointer to XMS handler                   */
WORD (*XMSHandler)(void);

/* DOS HMA and UMB functions                */
int  IsDOSHMAPresent(WORD *);
void far *DOSHMASubAlloc(WORD);
WORD SetStrategy(WORD);
WORD GetStrategy(void);
WORD SetUMBLink(void);
WORD AllocMem(WORD,WORD *);

/* DOS memory allocation strategies         */
#define ST_FIRST    (0x00)
#define ST_BEST     (0x01)
#define ST_LAST     (0x02)
#define ST_HIGH     (0x80)
#define ST_HIGHONLY (0x40)

#ifdef __BORLANDC__
    #pragma argsused
#endif
int main(int argc,char **argv)
{
   VideoTest();
   UMBTest();
   HMATest();
   return EXIT_SUCCESS;
}

void VideoTest (void)
{
   VideoNorm();

   /* Enable 128K of video RAM  */
   VideoRAM();
   VMsg(0,"Made it!");
   VMsg(1,"Use 0xA000:0x1000 as a pointer to a 124K"
          " data area.");
   VMsg(3,"Press a key");
   if(!getch())
      getch();
   VideoNorm();
}

void UMBTest (void)
{
   WORD wOldStrat,wWork;

   /* Now, a little UMB allocation  */
   wOldStrat=GetStrategy();
   if(!SetUMBLink()) {
      printf("Unable to set UMB link\n");
      return;
   }
   SetStrategy(ST_BEST|ST_HIGHONLY);

   /* 256 paragraphs == 4K  */

   if( AllocMem(256,&wWork)) { 
      printf("4K was allocated at %04x:0000\n",
         wWork);
   }
   else {
      printf("Unable to allocate 4K of UMB RAM\n");
      printf("%u paragraphs are available\n",
         wWork);
   }
   SetStrategy(wOldStrat);
}


void HMATest (void)
{
   WORD wWork;
   void far *lpWork;

   if(!InitXMS()) {
      printf("XMS not Available. Press a key\n");
      if(!getch())
         getch();
      return;
   }
   else {
      printf("XMS detected.\n");
   }

   if(XMSAllocHMA()) {
      printf("Accessing HMA via XMS\n");
      printf("Use address 0xFFF0:0x0010\n");
      XMSEnableA20();
      printf("A20 line is enabled\n");
      printf("Press a key\n");
      if(!getch())
         getch();
      XMSDisableA20();
      printf("A20 line is disabled\n");
      printf("Press a key\n");
      if(!getch())
         getch();
      return;      
   }   

   /* HMA Suballocation     */
   if(IsDOSHMAPresent(&wWork)) {
      printf("DOS is present in the HMA\n");
      printf("%u bytes are available\n");
      
      lpWork=DOSHMASubAlloc(wWork);
      if(lpWork==NULL) {
         printf("Unable to suballocate HMA\n");
      }
      else {
         printf("HMA block %p is allocated\n",
            lpWork);
      }
   }
   else {
      printf("DOS is not present in the HMA\n");
   }      
}

/* Enable 128K of RAM at 0xA000:0x0000  */
void VideoRAM (void)
{
   outp(0x3ce,6);
   outp(0x3cf,0);      
}

/* Send a message to relocated video line.  */
void VMsg (WORD wLine, char *msg )
{
   static unsigned char far *vid=
      (unsigned char far *) 0xa0000000L;

   int i;

   for(i=wLine*160;*msg;i+=2) {
      vid[i]=*msg++;
      vid[i+1]=0x07;
   }
}

/* Reset to 80-column color     */
void VideoNorm (void)
{
   regs.x.ax=3;
   int86(0x10,&regs,&regs);
}

/* Allocate RAM in paragraphs via DOS   */
WORD AllocMem ( WORD wParasToAlloc, WORD *wParas)
{
   regs.h.ah=0x48;
   regs.x.bx=wParasToAlloc;
   int86(0x21,&regs,&regs);
   if(regs.x.cflag & 1) { 
      *wParas=regs.x.bx;
      return(0);
   }
   else {
      *wParas=regs.x.ax;
      return(1);
   }
}

/* See if XMS is present and store XMS driver address. */
WORD InitXMS (void)
{
   regs.x.ax=0x4300;
   int86(0x2f,&regs,&regs);
   if(regs.h.al & 0x80) {
      // now get driver address
      regs.x.ax=0x4310;
      int86x(0x2f,&regs,&regs,&sregs);
      FP_SEG(XMSHandler)=sregs.es;
      FP_OFF(XMSHandler)=regs.x.bx;
      return(1);
   }
   else {
      return(0);
   }
}

/* Make a call to the XMS driver    */
WORD CallXMS ( WORD AXParm, WORD DXParm)
{
   _asm {
      push dx
      push bx
      mov  ax,AXParm
      mov  dx,DXParm
      call [XMSHandler]
      pop  bx
      pop  dx
   }
   /* AX has return code    */
}

/* Allocate entire HMA  */
WORD XMSAllocHMA (void)
{   return(CallXMS(01,0xffff));    }

/* Enable A20 Line  */
WORD XMSEnableA20(void)
{   return(CallXMS(03,0));   }

/* Disable A20 Line */
WORD XMSDisableA20(void)
{   return(CallXMS(04,0));   }

/* See if DOS is in HMA */
int IsDOSHMAPresent ( WORD *wFree)
{
   regs.x.ax=0x4a01;
   int86(0x2f,&regs,&regs);
   if(regs.x.bx==0) {
      return(0);
   }
   else {
      *wFree=regs.x.bx;
      return(1);
   }
}

/* Get ptr from first available DOS suballocation of HMA */
void far *DOSHMASubAlloc( WORD wBytes)
{
   struct SREGS sregs;
   void far *v;

   regs.x.ax=0x4a02;
   regs.x.bx=wBytes;

   int86x(0x2f,&regs,&regs,&sregs);
   FP_SEG(v)=sregs.es;
   FP_OFF(v)=regs.x.di;
   return( v ) ;   
}

/* Link upper-memory chain*/
WORD SetUMBLink (void)
{
   regs.x.ax=0x5803;
   regs.x.bx=1;
   int86(0x21,&regs,&regs);
   return(!(regs.x.cflag & 1));
}

/* Set DOS memory allocation strategy   */
WORD SetStrategy (WORD wStrat)
{
   regs.x.ax=0x5801;
   regs.x.bx=wStrat;
   int86(0x21,&regs,&regs);
   return(!( regs.x.cflag & 1));
}

/* Get current DOS memory allocation strategy   */
WORD GetStrategy (void)
{
   regs.x.ax=0x5800;
   int86(0x21,&regs,&regs);
   return(regs.x.ax);
}

